#!/usr/bin/python
"""Contains a class with methods to represent and operate on the product of
factors with exponents
"""
# split_code(), _format(), and _FORMATS are based on pint.formatting
# (https://github.com/hgrecco/pint, 7/11/14):
# - Copyright 2013 by Pint Authors
# - BSD license

# pylint: disable=I0011, C0103, E0213, R0904, W0142, W0223, W0621

from __future__ import division

import re

from collections import Counter
from fractions import Fraction
from .util import get_group, num2super

# Default keyword arguments sent to format()
_DEFAULT_FORMAT = dict(
    mul='*',
    div='/',
    group='({0})',
    base_fmt=lambda x: x,
    exp_fmt=lambda x: '(%s)' % x if isinstance(x, Fraction) else '%s' % x,
)

# Map format specifications to sets of keyword arguments for format.
_FORMATS = {
    '': _DEFAULT_FORMAT,
    'H':  # HTML
    dict(_DEFAULT_FORMAT,
         mul='&nbsp;',
         div=None,
         exp_fmt=lambda x: '<sup>%s</sup>' % x,
         ),
    'L':  # LaTeX
    dict(_DEFAULT_FORMAT,
         mul='\,',
         div=None,
         group=r'\left({0}\right)',
         base_fmt=lambda x: r'\mathrm{%s}' % x,
         exp_fmt=lambda x: ('^{%s}' % x
                             if isinstance(x, Fraction) or x < 0 else
                             '^%s' % x),
         ),
    'M':  # Modelica
    dict(_DEFAULT_FORMAT,
         mul='.',
         ),
    'U':  # Unicode
    dict(_DEFAULT_FORMAT,
         mul=' ',
         div=None,
         exp_fmt=num2super,
         ),
    'V':  # Verbose
    dict(_DEFAULT_FORMAT,
         mul=' * ',
         div=' / ',
         exp_fmt=lambda x: ('**(%s)' % x if isinstance(x, Fraction) else
                             '**%s' % x),
         ),
    }

_KNOWN_FORMATS = frozenset(_FORMATS)


def split_code(code):
    """Split a string format code into standard and exponent-specific parts.
    """
    # Based on pint.formatting (https://github.com/hgrecco/pint, 7/11/14).
    # Copyright 2013 by Pint Authors
    # License: BSD

    std_fmt = ''
    exp_fmt = ''
    for char in code:
        if char in _KNOWN_FORMATS:
            if exp_fmt:
                raise ValueError("Multiple exponent format codes found")
            exp_fmt = char
        else:
            std_fmt += char
    return std_fmt, exp_fmt


def _format(factors, **kwargs):
    """Format the product of (base, exponent) pairs as a string

    **Parameters:**

    - *factors*: List of (base, exponent) pairs

       Each base is a string that represents the base of a factor.  Each
       exponent is a number (:class:`numbers.Number` instance).

    - *mul*: String used to represent multiplication

    - *div*: String used to represent division

         If *div* is *None*, then negative powers are used instead of division.

    - *group*: Format used to group an expression

         This string must contain '{0}' as a placeholder for the grouped
         expression.  If *group* is *None*, the multiple division operators will
         be used if necessary.

    - *base_fmt*: Function that takes a base and formats it as a string

    - *exp_fmt*: Function that takes an exponent and formats it as a string
    """
    # pylint: disable=I0011, R0912

    if not factors:
        return ''

    # Process the arguments.
    mul = kwargs['mul']
    div = kwargs['div']
    group = kwargs['group']
    base_fmt = kwargs['base_fmt']
    if div:
        exp_fmt = lambda x: kwargs['exp_fmt'](abs(x))
    else:
        exp_fmt = kwargs['exp_fmt']

    # Generate lists of strings of the numerator and denominator terms.
    num_terms = []
    den_terms = []
    for b, e in sorted(factors):
        base = base_fmt(b)
        if e == 1:
            num_terms.append(base)
        elif e > 0:
            num_terms.append(base + exp_fmt(e))
        elif e == -1 and div:
            den_terms.append(base)
        else:
            den_terms.append(base + exp_fmt(e))

    if not div:
        # Show as product using negative exponents.
        return mul.join(num_terms + den_terms)

    # Show as ratio, where all factors have positive exponents.
    num = mul.join(num_terms) or '1'

    if not den_terms:
        return num

    if group:
        den = mul.join(den_terms)
        if len(den_terms) > 1:
            den = group.format(den)
    else:
        den = div.join(den_terms)

    return div.join([num, den])

# Basic regular expressions:
u = r'\d+'  # Unsigned integer
i = '[+-]?' + u  # Integer
fl1 = i + r'\.(?:' + u + ')?(?:[Ee]' + i + ')?' # Float due to decimal point
fl2 = i + r'(?:\.' + u + ')?(?:[Ee]' + i + ')'  # Float due to exponential
fl = '(?:%s)|(?:%s)' % (fl1, fl2)  # Float
fr = i + '/' + u  # Fraction

# Regular expression parser for a base and an exponent:
base = '([A-Za-z][A-Za-z_]*)'
exponent = r'(?:(?:\((%s)\))|(%s)|(%s))' % (fr, fl, i)
remainder = '(.*)'
parser = re.compile('%s%s?%s?' % (base, exponent, remainder))
del fl1, fl2, fl, fr, base, exponent, remainder


def _split(factor):
    """Split a factor into a base, an exponent, and a remainder.
    """
    # Parse the factor.
    base, exp_fr, exp_fl, exp_i, remainder = parser.match(factor).groups()

    # Cast the exponent into the appropriate number type.
    if exp_fr is not None:
        exp = Fraction(exp_fr)
    elif exp_fl is not None:
        exp = float(exp_fl)
    elif exp_i is not None:
        exp = int(exp_i)
    else:
        exp = 1

    return base, exp, remainder


class Exponents(Counter):

    r"""Dictionary-like class to track exponents of independent bases

    The keys are the bases and the values are the exponents.  The supported
    mathematical operations (addition, subtraction, negation, multiplication,
    and division) operate on the exponents, not on the bases.

    **Initialization signatures:**

    - :class:`Exponents`\(): Returns an :class:`Exponents` instance with an
      empty set of factors

    - :class:`Exponents`\(a=\ *exp1*, b=\ *exp2*, ...), where *exp1* and *exp2*
      are numbers: Returns an :class:`Exponents` instance that represents the
      product of symbol 'a' raised to the power of *exp1* and symbol 'b' raised
      to the power of *exp2*

    - :class:`Exponents`\(dict(a=\ *exp1*, b=\ *exp2*, ...)): Returns the same
      result as the previous signature

    - :class:`Exponents`\({a: *exp1*, b: *exp2*, ...}): Returns the same result
      as the previous signature

    - :class:`Exponents`\(*string*): Returns an :class:`Exponents` instance
      indicated by *string*

         *string* must follow the format accepted by the :meth:`fromstr`
         constructor.

    **Formatting**

    An :class:`Exponents` instance can be expressed using `string format syntax
    <https://docs.python.org/2/library/string.html#format-string-syntax>`_.  The
    format codes are:

    - '' (default):

         - Exponents directly follow the symbols of the bases.
         - '*' indicates multiplication and '/' indicates division.
         - If the denominator contains multiple factors, then they are grouped
           in parentheses.

    - 'H' (HTML):

         - Exponents are written as superscripts ('<sup>...</sup>')
         - A non-breaking space ('&nbsp;') indicates multiplication and '/'
           indicates division.
         - If the denominator contains multiple factors, then they are grouped
           in parentheses.

    - 'L' (LaTeX_ math):

         - Exponents are written as superscripts ('^...')
         - Back-to-back factors indicate multiplication and '/' indicates
           division.
         - If the denominator contains multiple factors, then they are grouped
           in parentheses.
         - The output must be typeset in LaTeX_ math mode (e.g., surround it
           with '$...$').

    - 'M' ([Modelica]_): Same as the default except '.' indicates multiplication
      instead of '*'

    - 'U' (Unicode_):

         - Exponents are indicated by Unicode_ superscripts.  Only integer
           exponents are supported.
         - A space indicates multiplication.  Negative exponents are used
           instead of division.

    - 'V' (verbose):

         - Exponents are noted by '**'.
         - ' * ' indicates multiplication and ' / ' indicates division.
         - If the denominator contains multiple factors, then they are grouped
           in parentheses.

    **Examples:**

    Initialization and representation:

    >>> x = Exponents(a=1, b=-1, c=-2)
    >>> dict(x)
    {'a': 1, 'b': -1, 'c': -2}
    >>> str(x)
    'a/(b*c2)'
    >>> format(x, 'V')
    'a / (b * c**2)'
    >>> format(x, 'H')
    'a&nbsp;b<sup>-1</sup>&nbsp;c<sup>-2</sup>'

    .. role:: raw-html(raw)
       :format: html

    The last result displays in HTML as the following:
    :raw-html:`a&nbsp;b<sup>-1</sup>&nbsp;c<sup>-2</sup>`.

    Addition:

    >>> print(x + x)
    a2/(b2*c4)

    Multiplication:

    >>> print(x*2)
    a2/(b2*c4)
    >>> str(0*x)
    ''

    Division:

    >>> print(x/2)
    a0.5/(b0.5*c)

    Negation:

    >>> print(-x)
    b*c2/a

    Subtraction:

    >>> str(x - x)
    ''

    Indexing:

    >>> x['a']
    1
    >>> x['d']
    0

    Updating:

    >>> x['a'] += 1
    >>> x['d'] += 2
    >>> x['b'] = 2
    >>> str(x)
    'a2*d2/(b2*c2)'


    .. _Unicode: https://en.wikipedia.org/wiki/Unicode
    .. _LaTeX: http://www.latex-project.org/

    .. rubric:: References

    .. [Modelica] Modelica Specification, version 3.3, p. 235--236
       (https://www.modelica.org/documents)
    """

    def update(self, *args, **kwargs):
        """update([E, ]**F) -> None

        Update this :class:`Exponents` instance from :class:`dict`/iterable *E*
        and *F*.

        If *E* is present, this does:

           ``for k in E: D[k] = E[k]``

        Then it does:

           ``for k in F: D[k] = F[k]``

        Alternatively, *E* can be a string that is compatible with the
        :meth:`fromstr` constructor.

        **Example:**

        >>> e = Exponents()
        >>> e.update(a=1)
        >>> e.update(dict(b=2), c=3)
        >>> e.update('1/d')
        >>> e # doctest: +SKIP
        Exponents({'a': 1, 'b': 2, 'c': 3, 'd': -1})

        .. testcleanup::
           >>> assert e == dict(a=1, b=2, c=3, d=-1)
        """
        try:
            # Assume args[0] is a string.
            arg = args[0].replace(' ', '')  # Remove spaces.
        except (IndexError, AttributeError):
            Counter.update(self, *args, **kwargs)
        else:
            if len(args) > 1:
                raise TypeError("update expected at most 1 arguments, got %i"
                                % len(args))
            Counter.update(self, Exponents.fromstr(arg), **kwargs)

    @classmethod
    def fromstr(cls, expr):
        """Create an :class:`Exponents` instance from a string expression.

        **Parameters:**

        - *expr*: String expression that indicates the symbols and exponents

             Each symbol may be followed directly by an exponent.  If the
             exponent is not an integer, it may be expressed using a decimal or
             as a fraction (with '/') enclosed in parentheses.  Factors may be
             multiplied('*'), divided ('/'), or grouped ('(...)').

        **Example:**

        >>> e = Exponents.fromstr('a/b/(c*d2)')
        >>> print(e)
        a/(b*c*d2)
        """
        # Note: It's not possible to use the ast module (at least not easily)
        # because a valid expression is not necessarily valid Python.  pyparser
        # is an option, but it's much slower than re.

        def process_expression(expr, exponents, add=True):
            """Process a factor from the left of a string expression (expr).

            Add (if add == True) the factor to exponents if add == True.
            Otherwise, subtract it.  Return the remainder of the expression.
            """
            # Pass if the expression is unity.
            if expr[0] == '1':
                return expr.strip('1')

            # Find the base, exponent, and remainder of the expression.
            base, exp, remainder = _split(expr)

            # Update the exponents.
            if add:
                exponents[base] += exp
            else:
                exponents[base] -= exp

            return remainder

        new = cls()
        add = True
        while len(expr) >= 1:
            if expr[0] == '(':
                # Recurse into the subexpression:
                subexpr, expr = get_group(expr)
                if add:
                    new += cls.fromstr(subexpr)
                else:
                    new -= cls.fromstr(subexpr)
            else:
                expr = process_expression(expr, new, add)

            # Get the operator for the next loop.
            if len(expr) >= 1:
                add = expr[0] == '*'
                assert add or expr[0] == '/', (
                    "The operation must be '*' or '/'.")
                expr = expr[1:]
                assert expr, "The expression can't end with an operator."

        return new

    def __add__(x, y):
        """x.__add__(y) <==> x+y"""
        copy = x.copy()
        for base, exp in y.items():
            copy[base] += exp
        return copy

    __radd__ = __add__
    __radd__.__doc__ = "x.__radd__(y) <==> y+x"

    def __iadd__(x, y):
        """x.__iadd__(y) <==> x+=y"""
        for base, exp in y.items():
            x[base] += exp
        return x

    def __sub__(x, y):
        """x.__sub__(y) <==> x-y"""
        result = x.copy()
        for base, exp in y.items():
            result[base] -= exp
        return result

    def __rsub__(x, y):
        """x.__rsub__(y) <==> y-x"""
        result = -x
        for base, exp in y.items():
            result[base] += exp
        return result

    def __isub__(x, y):
        """x.__isub__(y) <==> x-=y"""
        for base, exp in y.items():
            x[base] -= exp
        return x

    def __mul__(x, y):
        """x.__mul__(y) <==> x*y"""
        if y == 0:
            return x.__class__()
        return x.__class__({base: exp * y for base, exp in x.items()})

    __rmul__ = __mul__
    __rmul__.__doc__ = "x.__rmul__(y) <==> y*x"

    def __truediv__(x, y):
        """x.__truediv__(y) <==> x/y"""
        return x.__class__({base: exp / y for base, exp in x.items()})

    __div__ = __truediv__
    __div__.__doc__ = "x.__div__(y) <==> x/y"

    def __rtruediv__(x, y):
        """x.__rtruediv__(y) <==> y/x"""
        return x.__class__({base: y / exp for base, exp in x.items()})

    __rdiv__ = __rtruediv__
    __rdiv__.__doc__ = "x.__rdiv__(y) <==> y/x"

    def __imul__(x, y):
        """x.__imul__(y) <==> x*=y"""
        if y == 0:
            x.clear()
        else:
            for base, exp in x.items():
                x[base] = exp * y
        return x

    def __neg__(x):
        """x.__neg__() <==> -x"""
        return x.__class__({base: -exp for base, exp in x.items()})

    def __pos__(x):
        """x.__pos__() <==> x"""
        return x

    def __setitem__(x, i, y):
        """x.__setitem__(i, y) <==> x[i]=y"""
        if y == 0:
            del x[i]
        else:
            Counter.__setitem__(x, i, y)

    def __format__(self, format_code=''):
        """Format the Exponents instance according to format_code.
        """
        try:
            fmt = _FORMATS[format_code]
        except KeyError:
            raise ValueError("Unknown format code '%s' for object of type '%s'"
                             % (format_code, self.__class__.__name__))
        result = _format(self.items(), **fmt)
        return result

    def __str__(self):
        """Return an informal string representating the Exponents instance.
        """
        return format(self)



#!/usr/bin/python
"""Set up the natu package.

See README.md for instructions.
"""

# pylint: disable=C0103

import re

from os import path
from setuptools import setup

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'natu', '__init__.py')) as f:
    hit = re.search('__version__ *= *["'"'"'](.*)["'"'"']', f.read())
    try:
        version = hit.group(1)
    except AttributeError:
        version = None

with open(path.join(here, 'doc/long-description.txt')) as f:
    long_description = f.read()

setup(name='natu',
      version=version if version else '0',
      description="Natural units in Python",
      long_description=long_description,
      author='Kevin Davies',
      author_email='kdavies4@gmail.com',
      license='BSD-compatible (see LICENSE.txt)',
      keywords=('quantity calculus quantities unit conversion natural SI CGS '
                'Planck Hartree'),
      url='http://kdavies4.github.io/natu/',
      download_url=('https://github.com/kdavies4/natu/archive/v%s.zip'
                    % version if version else ''),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Education',
          'Topic :: Scientific/Engineering',
          'Topic :: Software Development :: Quality Assurance',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.2',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Operating System :: POSIX :: Linux',
          'Operating System :: Microsoft :: Windows',
      ],
      provides=['natu'],
      packages=['natu', 'natu.config', 'natu.groups'],
      package_data={'natu.config': ['*.ini']},
      platforms='any',
      zip_safe=False, # because ini files must be accessed
      test_suite = 'tests.test_suite',
     )



# Standard Library imports
import unittest

# Enthought Library imports
from traits.testing.api import doctest_for_module

# Numerical modeling library imports
from scimath.units import function_signature
from scimath.units.function_signature import (
    function_arguments, def_signature, call_signature,
)

class FunctionArgumentsDocTestCase(doctest_for_module(function_signature)):
    pass

class FunctionArgumentsTestCase(unittest.TestCase):

    ############################################################################
    # TestCase interface.
    ############################################################################

    def setUp(self):
        unittest.TestCase.setUp(self)

    def tearDown(self):
        unittest.TestCase.tearDown(self)


    ############################################################################
    # FunctionArgumentsTestCase interface.
    ############################################################################

    def test_single(self):
        """ Does a function with a single positional argument work?
        """
        def func(value):
            pass
        args, kw, arg_ordered = function_arguments(func)
        self.assertEqual(args,('value',))
        self.assertEqual(kw,{})
        self.assertEqual(arg_ordered,('value',))

    def test_multiple(self):
        """ Does a function with multiple positional argument work?
        """
        def func(a,b,c):
            pass
        args, kw, arg_ordered = function_arguments(func)
        self.assertEqual(args,('a','b','c'))
        self.assertEqual(kw,{})
        self.assertEqual(arg_ordered,('a','b','c'))

    def test_single_kw(self):
        """ Does a function with a single keyword argument work?
        """
        def func(a=1):
            pass
        args, kw, arg_ordered = function_arguments(func)
        self.assertEqual(args,())
        self.assertEqual(kw,{'a':1})
        self.assertEqual(arg_ordered,('a',))

    def test_multiple_kw(self):
        """ Does a function with multiple keyword arguments work?
        """
        def func(a=1,b=2):
            pass
        args, kw, arg_ordered = function_arguments(func)
        self.assertEqual(args,())
        self.assertEqual(kw,{'a':1,'b':2})
        self.assertEqual(arg_ordered,('a','b'))

    def test_single_arg_and_kw(self):
        """ Does a function with one positional and one keyword arg work?
        """
        def func(a, b=1):
            pass
        args, kw, arg_ordered = function_arguments(func)
        self.assertEqual(args,('a',))
        self.assertEqual(kw,{'b':1})
        self.assertEqual(arg_ordered,('a','b'))

    def test_multiple_arg_and_kw(self):
        """ Does a function with two positional and two keyword arg work?
        """
        def func(a, b, c=1,d=2):
            pass
        args, kw, arg_ordered = function_arguments(func)
        self.assertEqual(args,('a','b'))
        self.assertEqual(kw,{'c':1,'d':2})
        self.assertEqual(arg_ordered,('a','b','c','d'))


class DefSignatureTestCase(unittest.TestCase):

    ############################################################################
    # TestCase interface.
    ############################################################################

    def setUp(self):
        unittest.TestCase.setUp(self)

    def tearDown(self):
        unittest.TestCase.tearDown(self)


    ############################################################################
    # FunctionArgumentsTestCase interface.
    ############################################################################

    def test_kw(self):
        """ Does a function with a keyword and positional argument work?
        """
        def func(a,b=1):
            pass
        res = def_signature(func)
        self.assertEqual(res, 'def func(a, b=1):')

    def test_kw_with_list_argument(self):
        """ Does a function with a keyword argument as a list work?
        """
        def func(a,b=[]):
            pass
        res = def_signature(func)
        self.assertEqual(res, 'def func(a, b=[]):')

class CallSignatureTestCase(unittest.TestCase):

    ############################################################################
    # TestCase interface.
    ############################################################################

    def setUp(self):
        unittest.TestCase.setUp(self)

    def tearDown(self):
        unittest.TestCase.tearDown(self)


    ############################################################################
    # FunctionArgumentsTestCase interface.
    ############################################################################

    def test_kw(self):
        """ Does a function with a keyword and positional argument work?
        """
        def func(a,b=1):
            pass
        res = call_signature(func)
        self.assertEqual(res, 'func(a, b)')

    def test_kw_with_list_argument(self):
        """ Does a function with a keyword argument as a list work?
        """
        def func(a,b=[]):
            pass
        res = call_signature(func)
        self.assertEqual(res, 'func(a, b)')


# Some functions to test on.
def just_args(x, y):
    pass

def just_kwds(y=1, x=2):
    pass

def args_and_kwds(x, z=1, y=2):
    pass


class FunctionSignatureTestCase(unittest.TestCase):

    def test_just_args(self):
        self.assertEquals(def_signature(just_args), "def just_args(x, y):")

    def test_just_kwds(self):
        self.assertEquals(def_signature(just_kwds), "def just_kwds(y=1, x=2):")

    def test_args_and_kwds(self):
        self.assertEquals(def_signature(args_and_kwds), "def args_and_kwds(x, z=1, y=2):")

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
""":mod:`numpy`, adapted for use with physical quantities

Many of the functions only accept dimensionless quantities, and they operate on
the underlying values---not the values as represented in a particular display
unit.  To operate on values in a display unit would be to favor a particular
unit, which is against the `design of the package <index.html>`_.

The constants (*pi*, *e*) are exactly as they are in :mod:`numpy`.

These functions accept floats, integers, and dimensionless quantities:

- Hyperbolic functions: :func:`arccosh`, :func:`arcsinh`, :func:`arctanh`,
  :func:`cosh`, :func:`sinh`, and :func:`tanh`

They functions are the same as those from :mod:`numpy`.  The functions cast
quantities cast as floats and return floats.

These functions accept angle as a quantity:

- :func:`cos`, :func:`sin`, and :func:`tan`

These functions accept floats, integers, and dimensionless quantities:

- :func:`acos`, :func:`asin`, :func:`atan`, and :func:`atan2`

They return angles as quantities.

These functions are no longer applicable and have been deleted since angle is a
quantity:

- :func:`degrees`, :func:`radians`, :func:`rad2deg`, and :func:`deg2rad`

All other functions are directly imported from :mod:`numpy`.  However, some of
these need to be adapted (`Issue #7
<https://github.com/kdavies4/natu/issues/7>`_).
"""

from __future__ import absolute_import

from numpy import *
from . import _decorators as decor

# TODO: Update numpy.info for the modified functions.

# Trigonometric functions
# -----------------------

cos = decor.trig(cos)
sin = decor.trig(sin)
tan = decor.trig(tan)
arccos = decor.inv_trig(arccos)
arcsin = decor.inv_trig(arcsin)
arctan = decor.inv_trig(arctan)
#arctan2 = decor.inv_trig_yx(  # TODO fix
#    arctan2,
#    """Return the arc tangent of y/x as an angle.
#
#    Unlike atan(y/x), the signs of both x and y are considered.""")
#hypot = decor.homogeneous_copy_props(hypot)  # TODO fix

# These aren't needed since angle is a quantity:
del degrees, radians, rad2deg, deg2rad

#'unwrap'

# Hyperbolic functions
# --------------------

arccosh = decor.dimensionless(arccosh)
arcsinh = decor.dimensionless(arcsinh)
arctanh = decor.dimensionless(arctanh)
cosh = decor.dimensionless(cosh)
sinh = decor.dimensionless(sinh)
tanh = decor.dimensionless(tanh)

# TODO: Support the functions below.
# The ones as strings are from numpy.core.umath.
# The ones as comments are from elsewhere in numpy.
# Both can be imported from the base of the numpy module.

# Rounding
# --------

# around
# round_
#'rint'
# fix
#'floor'
#'ceil'
#'trunc'

# Sums, products, differences
# ---------------------------
# prod
# sum
# nansum
# cumprod
# cumsum
# diff
# ediff1d
# gradient
# cross # should introduce a factor of 1/cyc
# trapz

# Exponents and logarithms
# ------------------------

#'exp'
#'expm1'
#'exp2'
#'log'
#'log10'
#'log2'
#'log1p'
#'logaddexp'
#'logaddexp2'

# Other special functions
# -----------------------
# IO
# sinc

# Floating point routines
# -----------------------
#'signbit'
#'copysign'
#'frexp'
#'ldexp'

# Arithmetic operations
# ---------------------
#'add'
#'reciprocal'
#'negative'
#'multiply'
#'divide'
#'power'
#'subtract'
#'true_divide'
#'floor_divide'
#'fmod'
#'mod'
#'modf'
#'remainder'

# Handling complex numbers
# ------------------------
# angle
# real
# imag
#'conj'

# Miscellaneous
# -------------
# convolve
# clip
#'sqrt'
#'square'
#'absolute'
#'fabs'
#'sign'
#'maximum'
#'minimum'
#'fmax'
#'fmin'
# nan_to_num
# real_if_close
# interp

# Not on webpage:
#'bitwise_and'  # broken
#'bitwise_or'  # broken
#'bitwise_xor'  # broken
#'conjugate'  # broken
#'equal'  # works as is
#'euler_gamma'
#'frompyfunc'
#'geterrobj'
#'greater'
#'greater_equal'
#'invert'
#'isfinite'  # broken
#'isinf'  # broken
#'isnan'  # broken
#'left_shift'
#'less'
#'less_equal'
#'logical_and'
#'logical_not'
#'logical_or'
#'logical_xor'
#'nextafter'
#'not_equal'
#'right_shift'
#'seterrobj'
#'spacing'

del decor



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Greg Rogers
#
#-----------------------------------------------------------------------------

from unittest import TestCase

from traits.api import TraitError

from scimath.units.api import MetaQuantity


class TraitsTestCase(TestCase):

    def test_metaquantity(self):
        mq = MetaQuantity( name='vp', units='km/s', family_name='pvelocity')

        self.failUnlessEqual( mq.name, 'vp' )
        self.failUnlessEqual( mq.units.label, 'km/s' )
        self.failUnlessEqual( mq.family_name, 'pvelocity' )
        return

    def test_metaquantity_compatible_family_change(self):
        mq = MetaQuantity( name='vp', units='km/s', family_name='pvelocity')

        mq.name = 'vs'
        mq.family_name = 'svelocity'
        self.failUnlessEqual( mq.name, 'vs' )
        self.failUnlessEqual( mq.units.label, 'km/s' )
        self.failUnlessEqual( mq.family_name, 'svelocity' )
        return

    def test_metaquantity_compatible_units_change(self):
        mq = MetaQuantity( name='vp', units='km/s', family_name='pvelocity')

        mq.units = 'ft/s'
        self.failUnlessEqual( mq.name, 'vp' )
        self.failUnlessEqual( mq.units.label, 'ft/s' )
        self.failUnlessEqual( mq.family_name, 'pvelocity' )
        return

    def test_metaquantity_incompatible_units_change(self):
        mq = MetaQuantity( name='vp', units='km/s', family_name='pvelocity')

        self.failUnlessRaises(TraitError, setattr, mq, 'units', 'hours')

        self.failUnlessEqual( mq.name, 'vp' )
        self.failUnlessEqual( mq.units.label, 'km/s' )
        self.failUnlessEqual( mq.family_name, 'pvelocity' )
        return

    def test_metaquantity_incompatible_family_change(self):
        mq = MetaQuantity( name='vp', units='km/s', family_name='pvelocity')

        mq.family_name = 'time'

        self.failUnlessEqual( mq.name, 'vp' )
        self.failUnlessEqual( mq.units.label, 'msec' )
        self.failUnlessEqual( mq.family_name, 'time' )
        return

    def ui_simple(self):
        mq = MetaQuantity()
        mq.configure_traits(kind='modal')
        print '\n'
        mq.print_traits()
        return

### EOF



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Travis N. Vaught
# Date: 05/22/2005
# Description: Very basic 'unit' unit tests.
#
#------------------------------------------------------------------------------

#############################################################################
# Imports:
#############################################################################
# Standard library imports
import unittest
import logging

# Major package imports
import numpy

# Local units imports
import scimath.units as units
from scimath.units.mass import kg, metric_ton
from scimath.units.temperature import kelvin, celsius, fahrenheit
from scimath.units import area, density, speed, time, frequency, \
        acceleration, temperature, length
from scimath.units.quantity import Quantity
from scimath.units.style_manager import style_manager
from scimath.units.unit_manager import unit_manager
from scimath.units.smart_unit import is_dimensionless

from scimath.units.speed import meters_per_second

from scimath.units.unit_parser import unit_parser


logger = logging.getLogger(__name__)


#############################################################################
# Assignment:
#############################################################################

ms2 = acceleration.meters_per_second_squared
fs2 = acceleration.feet_per_second_squared

mass = 1000
acc = 9.8
acc_imp = 32.152231

converters = unit_manager.unit_converters
systems = unit_manager.unit_systems
test_data = numpy.arange(1,100, .10)

#############################################################################
# Tests:
#############################################################################

class test_units(unittest.TestCase):

    def test_with_si(self):
        f = self._compute_force(mass, acc)
        self.assertEqual(f, mass * acc)
        return

    def test_with_ton(self):
        f = self._compute_force(mass / 1000, acc, mass_units = metric_ton)
        self.assertEqual(f, mass * acc)
        return

    def test_with_imp(self):
        f = self._compute_force(mass, acc_imp, acc_units = fs2)
        self.assertAlmostEqual(f, mass * acc,4)
        return

    def test_temperature(self):
        k = 200
        c = -73.15
        f = -99.67
        k2c = units.convert(k, kelvin, celsius)
        self.assertAlmostEqual(k2c,c)
        k2f = units.convert(k, kelvin, fahrenheit)
        self.assertAlmostEqual(k2f,f)
        f2c = units.convert(f, fahrenheit, celsius)
        self.assertAlmostEqual(f2c,c)
        t1 = units.convert(-40.0, temperature.fahrenheit, temperature.celsius)
        self.assertAlmostEqual(t1, -40.0, 6)
        t2 = units.convert(0.0, temperature.celsius, temperature.fahrenheit)
        self.assertAlmostEqual(t2, 32.0, 6)
        return

    def test_area(self):
        self.assertEqual(1*area.square_mile/(640*area.acre),1.0)
        return

    def test_density(self):
        d1 = units.convert(1, density.lb_per_gal,
                           density.grams_per_cubic_centimeter)
        self.assertAlmostEqual(d1,0.1198284429,6)
        return

    def test_frequency(self):
        f1 = units.convert(1,frequency.hertz,frequency.khz)
        self.assertAlmostEqual(f1,0.001,6)
        return

    def test_time(self):
        t1 = units.convert(1, time.year, time.second)
        self.assertAlmostEqual(t1,(365.25*24*60*60),6)

    def test_speed(self):
        s1 = units.convert(55, speed.miles_per_hour, speed.meters_per_second)
        self.assertAlmostEqual(s1, 24.5872, 6)

    def test_unit_db(self):

        from scimath.units import unit_db
        udb = unit_db.UnitDB()
        udb.get_family_members_from_file()
        udb.get_unit_families_from_file()
        # spot check reading default data into db object
        self.assertEqual(udb.get_family_name("dt"),'psonic')


    def test_style_manager(self):

        test_quantity = Quantity(test_data, name="vp", units='m/s', family_name="pvelocity")

        self.assertEqual(style_manager.get_info(test_quantity, 'line'), 'solid')
        self.assertEqual(style_manager.get_info(test_quantity, 'color'), 'black')


    def test_style_manager_ranges (self):

        test_quantity = Quantity(test_data, name="vp", units='m/s', family_name="pvelocity")

        # unit_system optional, since kgs is default
        self.assertEqual(style_manager.get_range(test_quantity, unit_system='kgs'),
                                              (1.5, 6.5))
        self.assertEqual(style_manager.get_range(test_quantity.invert(), unit_system='kgs'),
                                              (500.0, 0.0))

    def test_conversion_tracks_parents(self):
        """ Test that a _converted_from traits is set appropriately.

        Quantity converted from another quantity via change_unit_system has
        the _converted_from trait set appropriately.
        """

        metric_system = unit_manager.get_unit_system('METRIC')
        metric_depth = metric_system.units('depth')

        q1 = Quantity(32, units='m', family_name='depth')
        q2 = q1.change_unit_system('IMPERIAL')
        q3 = q2.change_unit_system('METRIC')

        self.assertEqual(q1, q2._converted_from,
                         "Conversion failed to track conversion parent.")
        self.assertEqual(q2, q3._converted_from,
                         "Conversion failed to track conversion parent.")

        return

    def test_propagation_base(self):
        """ Tests that propagation works for an original quantity. """

        q1 = Quantity(10, units='m', family_name='depth')
        q1.propagate_data_changes()

        self.assertEqual(10, q1.data,
                         "Propagation modified the base data.")

        return

    def test_propagation(self):
        """ Tests data propagation for a single converted quantity. """

        q1 = Quantity(10, units='m', family_name='depth')
        q2 = q1.change_unit_system('IMPERIAL')
        q3 = q1.change_unit_system('METRIC')

        q2.data = 2 * q2.data
        q2.propagate_data_changes()

        self.assertAlmostEqual(20., q1.data, 1,
            "Propagation test expected data 20, got %s" % str(q1.data))

        q3.data = 3 * q3.data
        q3.propagate_data_changes()

        self.assertAlmostEqual(30., q1.data, 1,
            "Propagation test expected data 30, got %s" % str(q1.data))

        return

    def test_get_original(self):

        q1 = Quantity(10, units='m', family_name='depth')
        q2 = q1.change_unit_system('IMPERIAL')
        q3 = q1.change_unit_system('METRIC')

        self.assertEqual(q1, q1.get_original(),
            "Original quantity's get_original method failed to return self.")
        self.assertEqual(q1, q2.get_original(),
            "First child get_original failed to return original.")
        self.assertEqual(q1, q3.get_original(),
            "Second child get_original failed to return original.")

        return

    def test_unit_parser_dimensionless(self):
        dless = Quantity(1.0, units="", family_name="unknown")
        self.assert_( is_dimensionless(dless.units),
            "Dimensionless units not properly parsed")

    def test_unit_parser_caps(self):
        mpers_in_caps = Quantity(1.0, units="M/S", family_name="pvelocity")
        self.assertEqual(mpers_in_caps.units.derivation, meters_per_second.derivation,
            "Capitalized units rejected")


    def test_family_compatibility(self):
        """ test are_compatible_families """

        self.assert_( unit_manager.are_compatible_families( 'none', 'pvelocity'))
        self.assert_( unit_manager.are_compatible_families( 'none', 'foo'))
        self.assert_( unit_manager.are_compatible_families( 'foo', 'foo'))
        self.assert_( not unit_manager.are_compatible_families( 'bar', 'foo'))
        self.assert_( not unit_manager.are_compatible_families( 'bar', 'none'))


    def test_get_inverse(self):
        """ test get_inverse_family_name and get_inverse_name """

        # Try with family_names provided
        self.assertEqual (unit_manager.get_inverse_family_name('psonic'), 'pvelocity')
        self.assertEqual (unit_manager.get_inverse_name('psonic'), 'vp')
        # Now try with family_members provided
        self.assertEqual (unit_manager.get_inverse_family_name('dt'), 'pvelocity')
        self.assertEqual (unit_manager.get_inverse_name('dt'), 'vp')


    def test_unit_equal(self):
        """ test unit class __eq__ method"""
        um1 = length.m
        um2 = length.m
        uft = length.feet

        self.assert_( um1 == um2 )
        self.assert_( not( um1 == uft ))

        return

    def test_unit_notequal(self):
        """ test unit class __ne__ method"""
        um1 = length.m
        um2 = length.m
        uft = length.feet

        self.assert_( not( um1 != um2 ))
        self.assert_( um1 != uft )

        return

    def test_smart_unit_equal(self):
        """ test smart unit class __eq__ method"""
        qm1 = Quantity(1, units='m', family_name='depth')
        qm2 = Quantity(2, units='m', family_name='depth')
        qft = Quantity(1, units='feet', family_name='depth')

        self.assert_( qm1.units == qm2.units )
        self.assert_( not( qm1.units == qft.units ))

        return

    def test_smart_unit_notequal(self):
        """ test smart unit class __ne__ method"""
        qm1 = Quantity(1, units='m', family_name='depth')
        qm2 = Quantity(2, units='m', family_name='depth')
        qft = Quantity(1, units='feet', family_name='depth')

        self.assert_( not( qm1.units != qm2.units ))
        self.assert_( qm1.units != qft.units )

        return

    def test_get_family_name(self):
        """ test get_family_name with standard and ?/* type matching """

        self.assertEqual(unit_manager.get_family_name('svelo'), 'svelocity')
        self.assertEqual(unit_manager.get_family_name('vpsand'), 'pvelocity')

    def test_unit_system_units_call(self):
        """ test call to a unit_system's 'units' method """

        kgs = unit_manager.lookup_system('KGS')
        # do some funny stuff to get 'almost_equal' testing
        # first test 'value'
        self.assertAlmostEqual(kgs.units('rhog').value, unit_parser.parse_unit('1000*kg/m**3').value)
        self.assertAlmostEqual(kgs.units('pvelocity').value, unit_parser.parse_unit('1000*m/s').value)
        # now make sure derivations match
        self.assertEqual(kgs.units('rhog').derivation, unit_parser.parse_unit('1000*kg/m**3').derivation)
        self.assertEqual(kgs.units('pvelocity').derivation, unit_parser.parse_unit('1000*m/s').derivation)


    #########################################################################
    # Private Methods:
    #########################################################################

    def _compute_force(self, mass, acc, mass_units = kg,
                       acc_units = ms2, **unused_units):
        """
        """
        # algorithm units are
        mass_u = kg
        acc_u = ms2
        # convert into algorithm units
        mass = units.convert(mass, mass_units, mass_u)
        acc = units.convert(acc, acc_units, acc_u)
        # do the math :-)
        return mass * acc

### EOF #######################################################################
